#!/usr/bin/env python
# -*- coding: utf-8 -*-



import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gimpshelf import shelf
import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################
# -------------------------------------------------------------------------------------
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a Python plug-in for GIMP.
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools>Custom Font Layer Utilities'
#
# The first option creates new letter faces from over-sized visible layer/s
# The second option clears the contents of the old character layer before copying and pasting the new letter face
# The third and fourth options set the mode (legacy and new) to be used when copying and pasting to generate the new font face
# The fifth option removes non-font character layers from the image - apart from any layer name containing the word 'legend'
# The sixth option will re-position all layers to the origin of their parent image (Option 4 does this too)
# The seventh option will re-order the layers in the custom font file according to the following order
#		space.png (or  .ping)
#		Capital letters (A-Z)
#		Lower case letters (a-z)
#		Numbers (0-9)
#		Basic maths characters
# 		Punctuation characters
#		Other keyboard characters
#		Special characters (Pound Sterling sign, Copyright Sign, Registered Trademark sign)
# You can change this order by altering the appropriate lines (currently beginning at line no 130) of the plug-in
# The eighth option will display a message box listing all the characters (ie layers) defined in the custom font
# The ninth option defines the string to be used to append to single letter layer and/or path names
# The tenth option will add .png to all the layers with one character names ie  A to A.png
# The eleventh option will add .png to all the paths with one character names
# The twelfth option will combine the characters from the xcf custom font file selected, with the file selection tool below, with the current characters
# The thirteenth option is the xcf file selection tool
# The fourteenth option trims pixels from the character

# Version 0.01 Fixed adding .png to single letter layers and paths with special characters ie unicode characters
# Version 1.02 Adds an option to create a new face for a custom font from additional/decorative visible layers
# Version 1.03 Re-position space layer, if one exists, along with other letter layers and adds the option to remove non-font-character layers from the image
# Version 1.04 Trim and feather pixels from letter faces to remove halo effect
# Version 1.05 Fixed bug where character layer is not at image origin (and so doesn't copy new face correctly): Adds option to use different modes when copying from visible layers to font face
# Version 1.06 Added option to re-position all layers to the origin of their parent image ie with x and y co-ordinates set to zero: Re-organized dialog and filter structure so that all options can be set without any issues
# Version 1.07 Added option to clear the existing letter face before copying: Re-ordered flow of filter (yet again); Added options to combine two custom font files
# Version 1.08 Fixed potential bug when layer is bigger than image
#              Changed filter location
# Version 1.09 Added mode choice when combining two custom fonts
# Version 1.10 Added fixed position option for creating new characters from a visible layer
# Version 1.11 Adds an option to trim pixels from each character; changed plug-in name to Custom Font Layer Utilities
# Version 1.12 Adds an option to add lower-case characters to the custom font by copying and scaling, as a percentage, the upper-case characters (when the custom font has a path called base_line.png)
# Version 1.13 Adds an option to sandwich a character between a top layer and a bottom layer - optionally setting the layer size to the full width of the font image size
# Version 1.15 Return trim option inadvertently removed, add lower case now works with part fonts, layers containing the word "Legend" no longer removed from stack
# Version 1.16 Adds options for creating a display layer to save as a png file to act as a folder thumbnail image - shows name of custom font file and number of characters
# Version 1.17 Adds an option to customise a line of text to include in the display information
# Version 1.18 Fixes display of smaller sized thumbnails i.e. those smaller than 401 px in height
# Version 1.19 Adds an option to remove all paths from a custom font file
# Version 1.20 .png (or extension string) can now be added to two character path names
#              accepts unicode symbol characters
# Version 1.21 Added option to delete all layers from an image:
#              This speeds up the process of adding (kerning) paths from a custom font with kerning paths to another family member without the paths since there is currently no call to open paths from a file
# Version 1.22 Added option to rename layers as custom font layers
# Version 1.23 Fixes bug in adding lower-case characters
# Version 1.24 Adds options for increasing the layer space above and/or below each character for fonts that are still too close to the character upper and lower boundaries (can be useful before adding outer glow)
# Version 1.25 Adds option for adjusting the length of all kerning paths
# Version 1.26 Adds option for removing all layers
# Version 1.27 Adds option for combining paths 
# Version 1.28 Removed path options to a separate path utilities plug-in to reduce dialog height
# Version 1.29 Tabbed
#              Added tabs with re-ordered options to make the dialog fit smaller screen sizes
#              Made layer mode option available to top layer when merging top and bottom layers to current character



# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)



def do_stuff(image, layer,  new_mode, add_case, scale_amount, padding, do_sandwich, big_sandwich, do_display, display_credit, do_save_png,
            do_sort, do_place_at_origin, top_gap, bottom_gap, left_gap, right_gap, col_to_alpha,
            do_layer_info, do_force_custom_names, text_string, do_layer_names, do_remove_non_letter_layers, do_remove_all_letter_layers, trim, ptab):
    
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	
    kerning_adjustment = 0
    do_remove_paths = False
    do_path_names = False
    do_convert_paths = False
    do_new_letters = False
    xcffilename = "" 
    do_merge_files = False
    mode = fn_get_mode(new_mode)
	
    if do_sandwich:
        			
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_names_list = image.layers
        visi_loop, top, bottom, other_layers = 0,0,0,0
		
        while visi_loop < num_layers:
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "top": top += 1
            if visi_name == "bottom": bottom += 1 
            visi_loop += 1
        if top>0 or bottom>0: other_layers = 1

        
        visi_loop,first_layer = 0,-1
		
        while visi_loop < num_layers and other_layers > 0:
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if visi_name == "space.png" or visi_name == " .png": 
                has_a_space = True
                space_loop = visi_loop
            if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                if first_layer == -1: first_layer = visi_loop
                # turn off visibility of letter layers
                pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
            # turn off visibility of letter layers
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False)  
            visi_loop += 1
		
		
        visi_loop = 0
		
        while visi_loop < num_layers and other_layers > 0:
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
                
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
            #if len(visi_name) == 1 or len(visi_name) == 5 and ".png" in visi_name:
                layer_width = layer_names_list[visi_loop].width
                layer_height = layer_names_list[visi_loop].height
                offset_x, offset_y = pdb.gimp_drawable_offsets(layer_names_list[visi_loop])
                if big_sandwich:
                    pdb.gimp_layer_set_offsets(layer_names_list[visi_loop],(image.width-layer_width)/2,offset_y)
                if top>0:
                    # make a duplicate of the "top" layer and insert
                    top_id = pdb.gimp_image_get_layer_by_name(image,"top")
                    top_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                    layer_copy = pdb.gimp_layer_copy(top_id, True)
                    pdb.gimp_image_insert_layer(image, layer_copy, None, top_position)
                    pdb.gimp_layer_set_mode(layer_copy, mode)

                    pdb.gimp_item_set_visible(layer_copy,True)
                    # turn on layer visibility
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop],True) 
                    # Merge down the copied layer to the character layer
                    pdb.gimp_image_merge_down(image, layer_copy, 0)
                    active_layer = image.active_layer
                    # Re-set the id of the layer
                    layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
				
                if bottom>0:
                    pdb.gimp_item_set_visible(layer_names_list[visi_loop],True) 
                    bottom_id = pdb.gimp_image_get_layer_by_name(image,"bottom")
                    bottom_position = pdb.gimp_image_get_item_position(image, layer_names_list[visi_loop])
                    bottom_position += 1
                    layer_copy = pdb.gimp_layer_copy(bottom_id, True)
                    pdb.gimp_image_insert_layer(image, layer_copy, None, bottom_position)
                    # turn on layer visibility
                    pdb.gimp_item_set_visible(layer_copy,True) 
                    layer_names_list = image.layers
                    # Merge down to the bottom layer
                    pdb.gimp_layer_set_mode(layer_names_list[visi_loop], mode)
                    pdb.gimp_image_merge_down(image, layer_names_list[visi_loop], 0)
                    active_layer = image.active_layer
                    layer_names_list[visi_loop] = pdb.gimp_image_get_active_layer(image)
                    layer_names_list = image.layers
                    #pdb.gimp_layer_resize(layer_names_list[visi_loop], layer_width, layer_height, 0, 0)
                if not big_sandwich: 
                    pdb.gimp_layer_resize(layer_names_list[visi_loop], layer_width, layer_height,  0 - offset_x, 0 - offset_y)
				
                pdb.gimp_item_set_name(active_layer,visi_name)
                pdb.gimp_item_set_visible(active_layer,False) 
				
            visi_loop += 1
			
            
            pdb.gimp_item_set_visible(layer_names_list[first_layer],True)
            pdb.gimp_image_set_active_layer(image, layer_names_list[first_layer])

    if add_case:
        check_base_name = "base_line.png"
        check_base_line = pdb.gimp_image_get_vectors_by_name(image, check_base_name) #Check that a path describing the character baseline does exist
        if check_base_line < 0:
            check_base_name = "base-line"
            check_base_line = pdb.gimp_image_get_vectors_by_name(image, check_base_name) #Check alternative name for path describing the character baseline
        if check_base_line>=0:
            active_layer = pdb.gimp_image_get_active_layer(image) #Remember the active layer before we begin
            char_vector = pdb.gimp_image_get_vectors_by_name(image, check_base_name)		
            NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(char_vector)
            Char_Vector_Stroke = Char_Vector_StrokeIDs_List[0]
            TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(char_vector, Char_Vector_Stroke)
            base_line_y = ListOfControlPoints[1] # Extract first control point y position
            loop = 0 # create a loop to use with each upper case letter
            while loop<=25:
                letter = chr(65 + loop) # upper case
                letter_name = letter + ".png"
                new_letter = chr(97 + loop) # lower case
                new_letter_name = new_letter + ".png"
                new_letter_name_id = pdb.gimp_image_get_layer_by_name(image,new_letter_name) # check that a lower-case letter does not already exist
                letter_name_id = pdb.gimp_image_get_layer_by_name(image,letter_name) # check that an upper-case letter exists
                if not new_letter_name_id and letter_name_id:
                    letter_name_id = pdb.gimp_image_get_layer_by_name(image,letter_name) # get layer id for existing upper case letter
                    layer_position = pdb.gimp_image_get_item_position(image, letter_name_id) # get position of this layer in the layer stack
                    layer_copy = pdb.gimp_layer_copy(letter_name_id, True) # copy the layer
                    pdb.gimp_image_insert_layer(image, layer_copy, None, layer_position)  # and insert it in the layer stack
                    original_height = layer_copy.height # get height and width of the original u/c layer
                    original_width = layer_copy.width
                    scaled_height = float(original_height/100) * float(scale_amount) # get scale factors for height and width of layer
                    scaled_percent = float(original_width)/100.0
                    scaled_width = float((original_width)/100.0) * float(scale_amount)
                    layer_copy = pdb.gimp_item_transform_scale(layer_copy, 0, 0, scaled_width, scaled_height) # scale the layer to UI value
                    scaled_width += float (padding * 2.0) # update scaling to include any new layer padding size 
                    pdb.gimp_image_select_item(image, 2, layer_copy) # get content selection from alpha
                    saved_selection = pdb.gimp_selection_save(image) # save the selection (as a channel)
                    pdb.gimp_selection_none(image) # clear the selection
                    pdb.gimp_drawable_edit_clear(layer_copy) # clear the contents of the copied layer
                    pdb.gimp_image_select_item(image, 2, saved_selection) # re-load the selection from the channel
                    success, x0, y0, x1, y1 = pdb.gimp_selection_bounds(image) # get selection information
                    translate_amount_y = base_line_y - y1 # calculate how much to move the selection in the y direction
                    pdb.gimp_selection_none(image)
                    pdb.gimp_item_transform_translate(saved_selection, 0 + padding, translate_amount_y) # move the selection
                    pdb.gimp_image_select_item(image, 2, saved_selection) # re=load the selection
                    pdb.gimp_layer_scale(layer_copy, scaled_width, original_height, False) # scale the copied layer to match height of existing layers but with new scaled width
                    pdb.gimp_drawable_edit_fill(layer_copy, 2) # fill the selection with white
                    pdb.gimp_selection_none(image)
                    pdb.gimp_item_set_name(layer_copy, new_letter_name) # re-name the new layer to lower-case custom font format
                    pdb.gimp_item_set_visible(layer_copy, False) # turn off visibility
                    pdb.gimp_image_remove_channel(image, saved_selection) # remove unwanted save selection from channel list
				
                loop += 1
            pdb.gimp_image_set_active_layer(image, active_layer) # recall original active layer
			
    if do_force_custom_names:
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop, letter = 0, 65
        while loop < num_layers:
            if letter == 91: letter = 97
            if letter>122: letter = 65
            layer_name = chr(letter) + ".png"
            pdb.gimp_item_set_name(layer_list[loop],layer_name)
            letter += 1
            loop += 1
		
    if do_layer_names:
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop = 0
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_list[loop])
            layer_name = layer_name.decode('utf-8')
            if len(layer_name) == 1:
                layer_name += text_string
                pdb.gimp_item_set_name(layer_list[loop],layer_name)
            else:
                if layer_name == "space":
                    layer_name += text_string
                    pdb.gimp_item_set_name(layer_list[loop],layer_name)
            loop += 1
			
			
    if do_path_names:
        num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        loop = 0
        while loop < num_paths:
            path_name = pdb.gimp_item_get_name(vector_list[loop])
            path_name = path_name.decode('utf-8')
            if len(path_name) <3:
                path_name += text_string
                pdb.gimp_item_set_name(vector_list[loop],path_name)
            else:
                if path_name == "space":
                    path_name += text_string
                    pdb.gimp_item_set_name(vector_list[loop],path_name)
            loop += 1				
			
			
    if do_sort:
        space_case = [' ','space']
        upper_case = ['Z','Y','X','W','V','U','T','S','R','Q','P','O','N','M','L','K','J','I','H','G','F','E','D','C','B','A']
        lower_case = ['z','y','x','w','v','u','t','s','r','q','p','o','n','m','l','k','j','i','h','g','f','e','d','c','b','a']
        numerals = ['9','8','7','6','5','4','3','2','1','0']
        mathematical = ['-','+','=','/','*']
        punctuation = ['.',',',':',';','!','?','\'','\"']
        other_characters = ['$','%','^','&','_','{','}','@','#','~','|','\\','<','>','(',')','[',']','{','}']
        special_characters = unichr(163) + unichr(169) + unichr(174) # Pound Sterling sign, Copyright Sign, Registered Trademark sign
		
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_id_list = image.layers
        loop = 0
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_id_list[loop])
            if loop == 0:
                layer_name_list = [layer_name]
            else:
                layer_name_list.append(layer_name)
            loop += 1

        all_layers = layer_name_list
		
        for character in special_characters:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
								
        for character in other_characters:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
								
        for character in punctuation:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
						
        for character in mathematical:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
        for character in numerals:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
        for character in lower_case:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
        for character in upper_case:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
			
        for character in space_case:
            layer_test = character + ".png"
            if layer_test in all_layers:
                layer_id = pdb.gimp_image_get_layer_by_name(image,layer_test)
                pdb.gimp_image_raise_item_to_top(image,layer_id)
				
 
    if do_place_at_origin:
	
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
					
        loop = 0
        # Get the list of layer ids
        layer_names_list = image.layers
        
        while loop < num_layers:
            # make sure that existing layers are positioned at the origin of the image
            layer_names_list[loop].set_offsets(0, 0)
            loop += 1		


    #pdb.gimp_message("offsetting... top_gap = " + str(top_gap) )
    if top_gap>0 or bottom_gap>0:
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        #pdb.gimp_message("offsetting...")
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
                layer_id = layer_names_list[visi_loop]
                new_width = layer_id.width
                new_height =  layer_id.height + top_gap + bottom_gap
                old_offx, old_offy = pdb.gimp_drawable_offsets(layer_id)
                offx = 0
                offy = (top_gap + bottom_gap)
                #offy = old_offy - (top_gap + bottom_gap)
                #offy = 0
                #pdb.gimp_item_transform_translate(layer_id, offx, top_gap)
                #pdb.gimp_layer_resize(layer_id, new_width, new_height, offx, offy)
                pdb.gimp_layer_resize(layer_id, new_width, new_height, offx, top_gap)
                layer_id.set_offsets(0, 0)
			
            visi_loop += 1
        pdb.gimp_image_resize_to_layers(image)
		
		
    
    if left_gap<>0 or right_gap<>0:
        #pdb.gimp_message("offsetting... left_gap = " + str(left_gap) )
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        #pdb.gimp_message("offsetting...")
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
                layer_id = layer_names_list[visi_loop]
                new_width = layer_id.width + left_gap + right_gap
                new_height =  layer_id.height
				
			
                old_offx, old_offy = pdb.gimp_drawable_offsets(layer_id)
                offx = (left_gap)
                offy = 0
                #offy = old_offy - (top_gap + bottom_gap)
                #offy = 0
                #pdb.gimp_item_transform_translate(layer_id, offx, top_gap)
                #pdb.gimp_layer_resize(layer_id, new_width, new_height, offx, offy)
                pdb.gimp_layer_resize(layer_id, new_width, new_height, offx, offy)
                layer_id.set_offsets(0, 0)
			
            visi_loop += 1
        pdb.gimp_image_resize_to_layers(image)
			
    if do_remove_non_letter_layers or do_remove_all_letter_layers:
        
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        do_not_remove = "legend"
        
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            if do_remove_non_letter_layers:
                test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
                if test or "Legend" in visi_name or visi_name == "space.png" or do_not_remove.lower() in visi_name.lower():
			        # Re-position letter layers
                    layer_names_list[visi_loop].set_offsets(0, 0)
                else:
			        # Remove other layers
                    pdb.gimp_image_remove_layer(image, layer_names_list[visi_loop])
                    pdb.gimp_image_resize_to_layers(image)
            else:
			    # Remove all layers
                pdb.gimp_image_remove_layer(image, layer_names_list[visi_loop])
                
                    		
            visi_loop += 1 
			
    if trim > 0:
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
			    proc_trim_away(image, layer_names_list[visi_loop], trim)
                    		
            visi_loop += 1
			
    lights = False
    if lights:
        our_feather, our_threshold = 2.0, 0.3
        feather_on = pdb.gimp_context_get_feather()
        feather_x, feather_y = pdb.gimp_context_get_feather_radius()
        pdb.gimp_context_set_feather_radius(our_feather, our_feather)
        pdb.gimp_context_set_feather(True)
        threshold = pdb.gimp_context_get_sample_threshold()
        pdb.gimp_context_set_sample_threshold(our_threshold)
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test:
                custom_layer = layer_names_list[visi_loop]
                pdb.gimp_image_select_color(image , 2, custom_layer, white)
                success = pdb.gimp_edit_copy(custom_layer)
                if success:
                    is_empty = pdb.gimp_selection_is_empty(image)
                    if not is_empty:
					
					
                        temp_layer = gimp.Layer(image, "Temp Layer", custom_layer.width, custom_layer.height, RGBA_IMAGE, 100, 0) # layer 
                        image.add_layer(temp_layer, -1)
                        floating_sel = pdb.gimp_edit_paste(temp_layer, False)
                        #success = pdb.gimp_edit_copy(custom_layer)
                        # set the layer mode as per the dialog UI when the contents of the original character layer have not been cleared
                        pdb.gimp_floating_sel_anchor(floating_sel)
					
                        #pdb.gimp_selection_none(image)
                        pdb.gimp_edit_clear(custom_layer)
                        layer_names_list[visi_loop] = pdb.gimp_image_merge_down(image, temp_layer, 2)
                        #visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                        #visi_name = visi_name.decode('utf-8')
                        #floating_sel = pdb.gimp_edit_paste(layer_names_list[visi_loop], True)
                        #pdb.gimp_floating_sel_anchor(floating_sel)
                    		
            visi_loop += 1
			
        
        pdb.gimp_context_set_sample_threshold(threshold)
        pdb.gimp_context_set_feather(feather_on)
        pdb.gimp_context_set_feather_radius(feather_x, feather_y)


    if do_convert_paths:
        
        # Since there is no pdb to merge two paths together we need to 
        # deal with character path files that use two character parts i.e. one ending in char that acts as the base character 
        # and one ending in part by joining the part vector onto the char vector
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
		
        # create a list of the actual vector names
        loop, part_flag = 0, False
		
        while loop<len(vector_list):
            pdb.gimp_item_set_visible(vector_list[loop], False)
            vector_name = pdb.gimp_item_get_name(vector_list[loop])
            vector_name = vector_name.decode('utf-8') # in case of unicode characters 
            if "part" in vector_name: part_flag = True
            if loop == 0:
                vector_names_list = [vector_name]
            else:
                vector_names_list.append(vector_name)
				
            loop += 1
			
        loop = 0
        while loop<len(vector_list):
            vector_name = vector_names_list[loop]
            vector_id = vector_list[loop]
            if (len(vector_name) == 6 and " char" in vector_name): #check that the vector is one we can use
                next_vector_name = "none"
                if loop < (len(vector_list)-1):
                    next_vector_name = vector_names_list[loop+1]
                    #pdb.gimp_message(next_vector_name)
                    if (len(next_vector_name) == 6 and " part" in next_vector_name): # look to see if the next vector needs joining onto the it
                        part_vector_id = vector_list[loop + 1]
                        # get paths etc associated with the part vector
                        num_part_strokes, part_vector_list = pdb.gimp_vectors_get_strokes(part_vector_id)
                        stroke_loop = 0
                            
                        while stroke_loop<num_part_strokes:
                            # loop through paths that make up the part vector
                            part_vector_stroke = part_vector_list[stroke_loop]
                            # get information associated with the current vector part
                            TypeOfStrokeInPart_Vector,NumberOfPointsInPart_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(part_vector_id, part_vector_stroke)
                            part_ListOfControlPoints = ListOfControlPoints
                            # add current part path details to the base character
                            vector_id = pdb.gimp_vectors_stroke_new_from_points(vector_list[loop], 0, NumberOfPointsInPart_Vector, ListOfControlPoints, IsStrokeClosed)
							
                            stroke_loop += 1
				
				
                        pdb.gimp_image_remove_vectors(image,  vector_list[loop+1])
            loop += 1
			
    
			
    if do_remove_paths:
			
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        loop = 0
        while loop<num_vectors:
            vector_id = vector_list[loop]
            #active_vector_name = pdb.gimp_item_get_name(vector_id)
            #active_vector_name = active_vector_name.decode('utf-8')
            pdb.gimp_image_remove_vectors(image, vector_list[loop])
            loop += 1
			
			

    if kerning_adjustment<>0:
			
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
        loop = 0
        while loop<num_vectors:
		
            # look for recognised kerning path
            kerning_path_name = pdb.gimp_item_get_name(vector_list[loop])
            kerning_path_name = kerning_path_name.decode('utf-8')
			
            if (len(kerning_path_name) == 5 and ".png" in kerning_path_name) or (len(kerning_path_name) == 6 and ".png" in kerning_path_name):
                    
		
                #vector_id = vector_list[loop]
                
                kerning_vector = vector_list[loop]
                num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(kerning_vector)
                type, num_points, controlpoints, closed  = pdb.gimp_vectors_stroke_get_points(kerning_vector, stroke_ids[0])
                kerning_target_width = controlpoints[6] + kerning_adjustment
                v_position = pdb.gimp_image_get_vectors_position(image, kerning_vector)
                #if kerning_path_name == "AV.png":
                #    old_v = controlpoints[6]
                #    new_v = controlpoints[6] + kerning_adjustment
                #    pdb.gimp_message("old_v = " + str(old_v) + " : " + "new_v = " + str(new_v) )
                pdb.gimp_image_remove_vectors(image, kerning_vector)
                new_vector_id = pdb.gimp_vectors_new(image, kerning_path_name)
                pdb.gimp_image_insert_vectors(image, new_vector_id, None, v_position)
                points1 = [controlpoints[0],controlpoints[1]] * 3
                points2 = [kerning_target_width,controlpoints[7]] * 3
                points = points1 + points2
                #pdb.gimp_message(points)
                new_vector_id = pdb.gimp_vectors_stroke_new_from_points(new_vector_id ,0, len(points), points, False)
                #pdb.gimp_image_lower_item_to_bottom(image, kerning_vector)
                #pdb.gimp_item_set_name(new_vector_id, pdb.gimp_item_get_name(kerning_list[kerning_path_loop]))
			
			
            #            #kerning_vector = kerning_vector_ids[kerning_path_loop]
            #            num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(kerning_vector)                             
            #            type, num_points, controlpoints, closed  = pdb.gimp_vectors_stroke_get_points(kerning_vector, stroke_ids[0])
            #            kerning_target_width = controlpoints[6] + kerning_path_adjustment
            #            v_position = pdb.gimp_image_get_vectors_position(image, kerning_vector)
            #            pdb.gimp_image_remove_vectors(image, kerning_vector)
            #            new_vector_id = pdb.gimp_vectors_new(image, kerning_path_name)
            #            pdb.gimp_image_insert_vectors(image, new_vector_id, None, v_position)
            #            points1 = [controlpoints[0],controlpoints[1]] * 3
            #            points2 = [kerning_target_width,controlpoints[7]] * 3
            #            points = points1 + points2
            #            #pdb.gimp_message(points)
            #            new_vector_id = pdb.gimp_vectors_stroke_new_from_points(new_vector_id ,0, len(points), points, False)
            #            #pdb.gimp_image_lower_item_to_bottom(image, kerning_vector)
            #            #pdb.gimp_item_set_name(new_vector_id, pdb.gimp_item_get_name(kerning_list[kerning_path_loop]))
			
            #active_vector_name = pdb.gimp_item_get_name(vector_id)
            #active_vector_name = active_vector_name.decode('utf-8')
            #pdb.gimp_image_remove_vectors(image, vector_list[loop])
            loop += 1
			

    if do_layer_info:
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop = 0
        header = "This file contains the following custom font character/s:\n\n"
        layer_info = header
		
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_list[loop])
            layer_name = layer_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_list[loop], layer_name)
            if test:
                if len(layer_name) == 5:
                    layer_info += layer_name[0]
                elif len(layer_name) == 6:
                    pos = layer_name.find(".png")
                    layer_info += layer_name[0:pos]
            elif (layer_name == "space.png"):
                layer_info += "space"
		
            loop += 1
        layer_info = layer_info.decode('utf-8') # Allows unicode characters above 127 to be displayed without creating errors
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(layer_info)
        pdb.gimp_message_set_handler(old_handler)
		
		
    if col_to_alpha:
        alpha_col = pdb.gimp_context_get_foreground()
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        layer_list = image.layers
        loop = 0
		
        while loop < num_layers:
            layer_name = pdb.gimp_item_get_name(layer_list[loop])
            layer_name = layer_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_list[loop], layer_name)
            if test:
                pdb.plug_in_colortoalpha(image, layer_list[loop], alpha_col)
			
		
            loop += 1
	
	
    if do_display or do_save_png:
        filename = pdb.gimp_image_get_filename(image)
        if str(filename) == "None":
            do_display, do_save_png = False, False
            old_handler = pdb.gimp_message_get_handler()	
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message("A display layer can only be created for a file that has been saved. Please save this custom font file before trying again.")
            pdb.gimp_message_set_handler(old_handler)
				
    if do_display or do_save_png:
        upper_case, lower_case, number_case, punctuation_case = "ABCDEFGHIJKLMNOPQRSTUVWXYZ", "abcdefghijklmnopqrstuvwxyz", "0123456789", ".,:;'!?\""
        display_id = pdb.gimp_image_get_layer_by_name(image,"A.png") #Check for uppercase "A" - if it doesn't exist we do not carry on
        if display_id: 
            filename = pdb.gimp_image_get_filename(image) # get information about the image file name and file path
            file_extension = os.path.splitext(filename)[1]
            file_path = os.path.splitext(filename)[0]
            base_path = os.path.basename(filename)
            #pdb.gimp_message("Filename = " + str(base_path))
            visi_loop, ctr = 0,0
            num_layers, layer_ids = pdb.gimp_image_get_layers(image)
            layer_names_list = image.layers
            while visi_loop < num_layers:
                # turn off visibility of layers
                pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) 
                visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
                visi_name = visi_name.decode('utf-8')
                test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
                if not test:
                    if visi_name.lower() <> "space.png":
                        ctr += 1	
                visi_loop += 1
				
            no_of_chars = num_layers - ctr # number of characters, including space character, contained in the layer stack
            pdb.gimp_image_set_active_layer(image, layer_names_list[0])
			
            #count the number of kerning paths in the file
            num_paths, vector_ids = pdb.gimp_image_get_vectors(image)
            vector_list = image.vectors
            loop, kern_ctr = 0,0
            while loop < num_paths:
                path_name = pdb.gimp_item_get_name(vector_list[loop])
                path_name = path_name.decode('utf-8')
                if (len(path_name) == 5 and path_name.find(".png") == 1) or (len(path_name) == 6 and path_name.find(".png") == 2):
                    kern_ctr += 1
                loop += 1				
			
			
            # Make a copy of uppercase "A" and insert it in the layer stack
            display_copy = pdb.gimp_layer_copy(display_id, True)
            pdb.gimp_image_insert_layer(image, display_copy, None, -1)
            # Rename the new layer
            pdb.gimp_item_set_name(display_copy,"Display")
            # turn on visibility of layer
            pdb.gimp_item_set_visible(display_copy, True)
			
            display_offset = display_copy.width # make a note of the width of the uppercase "A" layer
            display_id2 = pdb.gimp_image_get_layer_by_name(image,"b.png") # Make a copy of lowercase "b" if present in the stack
            if display_id2:
                display_copy2 = pdb.gimp_layer_copy(display_id2, True)
                pdb.gimp_image_insert_layer(image, display_copy2, None, -1)
            else:
                display_id2 = pdb.gimp_image_get_layer_by_name(image,"B.png") # if lowercase "b" is not present use uppercase "B" or uppercase "A"
                if display_id2:
                    display_copy2 = pdb.gimp_layer_copy(display_id2, True)
                    pdb.gimp_image_insert_layer(image, display_copy2, None, -1)
                else:
                    display_copy2 = pdb.gimp_layer_copy(display_id, True)
                    pdb.gimp_image_insert_layer(image, display_copy2, None, -1)
            
            pdb.gimp_layer_set_offsets(display_copy2, display_offset, 0) # position the second display character layer
            # turn on visibility of layer
            pdb.gimp_item_set_visible(display_copy2, True)
            # merge the two characters
            display_id = pdb.gimp_image_merge_down(image, display_copy2, 0)
            # resize the image if required
            if display_id.width > image.width: pdb.gimp_image_resize(image, display_id.width, image.height, 0, 0)
            # get the text to put into the new text layer
            our_text = base_path + "\n" + str(no_of_chars) + " characters."
            if kern_ctr == 1:
                our_text += "\n" + str(kern_ctr) + " kerning path."
            else:
                our_text += "\n" + str(kern_ctr) + " kerning paths."
            if str(display_credit) <> "": our_text += "\n" + display_credit
            # add the new text layer to the layer stack
            text_layer = pdb.gimp_text_layer_new(image,our_text, "Sen", 24, 0)
            pdb.gimp_image_insert_layer(image, text_layer, None, -1)
            pdb.gimp_text_layer_set_color(text_layer, black)
            pdb.gimp_item_set_visible(text_layer, True)
            # sort out positioning for text if user text is present
            if str(display_credit) <> "": 
                if image.height < 401:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height)
                else:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height - 120)
            else:
                if image.height < 401:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height)
                else:
                    pdb.gimp_layer_set_offsets(text_layer, 8, image.height - 80)
            # create a copy of the text layer to fill as a text background
            text_copy = pdb.gimp_layer_copy(text_layer, True)
            pdb.gimp_image_insert_layer(image, text_copy, None, -1)
            pdb.gimp_drawable_fill(text_layer, 2) # white fill
            # merge the text layers into the display layer
            text_copy = pdb.gimp_image_merge_down(image, text_copy, 0)
            display_id = pdb.gimp_image_merge_down(image, text_copy, 0)
            pdb.gimp_layer_set_offsets(display_id, 0, 0)
            pdb.gimp_image_resize(image, display_id.width, display_id.height, 0, 0)
			
            if do_save_png:
                # save a png file without asking the user
                save_file = file_path + ".png"
                pdb.file_png_save2(image, display_id, save_file, save_file, False, 9, True, False, False, True, True, True, True)

	

    pdb.gimp_displays_flush()     
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
	
######################################################################################################################################

def fn_get_mode(new_mode):
        # Use Legacy Modes if new Default Modes are set to Normal Mode
        temp_mode = 0
        if new_mode <> 0:
            if new_mode == 1:
                temp_mode = 1
            elif new_mode == 2:
                temp_mode = 57
            elif new_mode == 3:
                temp_mode = 59
            elif new_mode == 4:
                temp_mode = 60
            elif new_mode == 5:
                temp_mode = 36
            elif new_mode == 6:
                temp_mode = 55
            elif new_mode == 7:
                temp_mode = 31
            elif new_mode == 8:
                temp_mode = 42
            elif new_mode == 9:
                temp_mode = 33
            elif new_mode == 10:
                temp_mode = 35
            elif new_mode == 11:
                temp_mode = 54
            elif new_mode == 12:
                temp_mode = 30
            elif new_mode == 13:
                temp_mode = 43
            elif new_mode == 14:
                temp_mode = 53
            elif new_mode == 15:
                temp_mode = 23
            elif new_mode == 16:
                temp_mode = 45
            elif new_mode == 17:
                temp_mode = 44
            elif new_mode == 18:
                temp_mode = 48
            elif new_mode == 19:
                temp_mode = 49
            elif new_mode == 20:
                temp_mode = 50
            elif new_mode == 21:
                temp_mode = 51
            elif new_mode == 22:
                temp_mode = 32
            elif new_mode == 23:
                temp_mode = 52
            elif new_mode == 24:
                temp_mode = 34
            elif new_mode == 25:
                temp_mode = 46
            elif new_mode == 26:
                temp_mode = 47
            elif new_mode == 27:
                temp_mode = 41
            elif new_mode == 28:
                temp_mode = 37
            elif new_mode == 29:
                temp_mode = 38
            elif new_mode == 30:
                temp_mode = 39
            elif new_mode == 31:
                temp_mode = 40
            elif new_mode == 32:
                temp_mode = 24
            elif new_mode == 33:
                temp_mode = 25
            elif new_mode == 34:
                temp_mode = 26
            elif new_mode == 35:
                temp_mode = 27
            elif new_mode == 36:
                temp_mode = 56
        mode = temp_mode
			

        return mode
	
	
######################################################################################################################################

def proc_trim_away(img, layer, amount):
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
######################################################################################################################################


def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Layer_Utils_v1-29_T",
    "Provides options for turning paths into character layers ",
    "Provides options for turning paths into character layers ",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Layer Utils v.1.29 Tabbed...",
    "RGBA",
    [
    # 1st Tab stroke options creating characters
    (PF_OPTION, "mode_option2", "New Mode to use with font face:", 0, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance")),
    (PF_TOGGLE, "add_case", "Create whitebase lower-case\n (requires 'base-line' path): ", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_SPINNER, "scale_letter", "Scale Character Amount: ", 75,(25,95,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pad_scaling", "Padding size for scaled letters (px): ", 0,(0,100,1),{'newLine':'1','tab':0}),
    (PF_TOGGLE, "sandwich_option", "Sandwich characters between \nlayers called top and bottom:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_TOGGLE, "restrict_option", "Set sandwich layer \nto canvas size:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_TOGGLE, "dopng", "Add a display layer:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_STRING, "pf_credit", "Display credit: ", "Decoration by a Gimp Lover.", 0,{'newLine':'1','tab':0}),
    (PF_TOGGLE, "save_png", "Save display layer as png image:", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
	
    # 2nd Tab Order Position Spacing options
    (PF_TOGGLE, "sort_option", "Sort layers into order:", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "origin_option", "Position layers to the origin:", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_top", "Add space to the top of all layers: ", 0,(0,100,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_bottom", "Add space to the bottom of all layers: ", 0,(0,100,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_left", "Adjust space to the left of all layers: ", 0,(-60,100,1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_right", "Adjust space to the right of all layers: ", 0,(-60,100,1),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "col_to_alph", "FG colour to transparency: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
	
	
    # 3rd Tab Naming, Removal, Trim
	
    (PF_TOGGLE, "info_option", "Display current characters \nin a message box:", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "rename_custom", "Force custom font names on layers:", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_STRING, "word", "Label String:", ".png", 0,{'newLine':'1','tab':2}),
    (PF_TOGGLE, "layer_option", "Add Label String to \nsingle letter layer labels:", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "remove", "Remove all non-font layers", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "remove_all_layers", "Remove ALL LAYERS from the image:", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_SPINNER, "pf_trim", "Trim pixels from character: ", 0,(0,50,1), {'newLine':'1','tab':2}),
	
	
	
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Character additions'),_('Order, Position, Spacing'),_('Naming, Removal, Trim')) ),
    ],
    [],
    do_stuff)

main()
